#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

"""Time interval domain (alias of `ScalarInterval`)."""

from ._scalar import ScalarInterval


TimeInterval = ScalarInterval

__all__ = ["TimeInterval"]
